# The software in this package is distributed under the GNU General
# Public License version 2 (with a special exception described below).
#
# A copy of GNU General Public License (GPL) is included in this distribution,
# in the file COPYING.GPL.
#
# As a special exception, if other files instantiate templates or use macros
# or inline functions from this file, or you compile this file and link it
# with other works to produce a work based on this file, this file
# does not by itself cause the resulting work to be covered
# by the GNU General Public License.
#
# However the source code for this file must still be made available
# in accordance with section (3) of the GNU General Public License.
#
# This exception does not invalidate any other reasons why a work based
# on this file might be covered by the GNU General Public License.

""" Helpers for logging

ShortLevelFormatter: provide a 4-char-sized field "shortlevel" for message
urgency (dbug/info/warn/err /crit)

I2nLogger: logger that provides a notice(), allows omission for str.format

Further ideas: ::
* allow milliseconds in dateformat field (see test_short_level_format)
* create own basicConfig-like function that uses our classes as default

.. codeauthor:: Christian Herdtweck, christian.herdtweck@intra2net.com
"""

import logging
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL, NOTSET

#: log level half-way between INFO and WARNING
NOTICE = (INFO + WARNING)/2
logging.addLevelName(NOTICE, 'NOTICE')

#: default formatting string for ShortLevelFormatter
DEFAULT_SHORT_LEVEL_FORMAT = '%(asctime)s:%(msecs)03d %(shortlevel)s| %(msg)s'

#: default formatting string for date/time in ShortLevelFormatter
DEFAULT_SHORT_LEVEL_DATE_FORMAT = '%H:%M:%S'

#: mapping from level name to level int for I2nLogger's set_level
LEVEL_DICT = dict(notset=NOTSET, debug=DEBUG, info=INFO, notice=NOTICE,
                  warning=WARNING, error=ERROR, critical=CRITICAL)

#: min level allowed in I2nLogger
MIN_LEVEL = NOTSET

#: max level allowed in I2nLogger
MAX_LEVEL = CRITICAL

class ShortLevelFormatter(logging.Formatter):
    """
    Formatter for logging handlers that allows use of format field "shortlevel"

    using this formatter, you can specify in the log message format string the
    field "shortlevel" which will introduce in your log messages a 4-char field
    for the log record urgency: "DEBUG" --> "dbug", "INFO" --> "info",
    "WARNING" --> "warn", "ERROR" --> "err ", "CRITICAL" --> "crit"

    All other functionality (like other format fields) is inherited from base
    class Formatter

    Usage example::

        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        formatter = ShortLevelFormatter('%(shortlevel)s| %(msg)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    You can even add new levels::

        notice = (logging.INFO + logging.WARNING)/2
        formatter.add_level(notice, 'note')
        logger.log(notice, 'more important than info but no warning nor error')

    .. seealso:: testing funcion :py:func:`test_short_level_format`
    """

    def __init__(self, fmt=DEFAULT_SHORT_LEVEL_FORMAT,
                 datefmt=DEFAULT_SHORT_LEVEL_DATE_FORMAT, *args, **kwargs):
        """ creates a new ShortLevelFormatter

        forwards all args to super class Formatter and initializes dict with
        levelno and short string representations
        """
        self.parent = super(ShortLevelFormatter, self)
        self.parent.__init__(fmt=fmt, datefmt=datefmt, *args, **kwargs)
        self._shortlevel_dict = {DEBUG: 'dbug', INFO: 'info', WARNING: 'warn',
                                 ERROR: 'err ', CRITICAL: 'crit',
                                 NOTSET: '----'}

    def format(self, record):
        """ create string representation of given log record """
        try:
            record.shortlevel = self._shortlevel_dict[record.levelno]
        except KeyError:
            record.shortlevel = '????'

        return self.parent.format(record)

    def add_level(self, levelno, shortlevel_str):
        """ add a new message urgency level

        :param int levelno: numeric urgency level
        :param str shortlevel_str: string representation of message urgency;
                                   should be of length 4
        :returns: nothing
        """

        self._shortlevel_dict[levelno] = shortlevel_str


class I2nLogger:
    """ a more convenient logger

    Features::
    * can be used as follows::

        logger.info('the result is {0}', result)    # more convenient than...
        logger.info('the result is {0}'.format(result))    # ... the original

    * Has a :py:meth:`Logger.notice` function
    * Has by default as only handler a :py:class:`logging.StreamHandler` to
      stdout with a :py:class:`ShortLevelFormatter`
    * Simplifies setting of logger's logging level using
      :py:meth:`Logger.set_level`. The level is kept in sync between logger and
      handlers and can be queried using :py:meth:`get_level`
    * can specify name and level and [date]fmt all in constructor

    ..note:: Creating multiple instances with the same name is possible but
             will internally use the same logger. Each instance overwrites at
             construction time the previous instances' handlers and levels!
    """

    def __init__(self, name, level=INFO, fmt=DEFAULT_SHORT_LEVEL_FORMAT,
                 datefmt=DEFAULT_SHORT_LEVEL_DATE_FORMAT)
        """ creates a I2nLogger; forwards args to logging.getLogger 
        
        ..note:: You should use a different name in each constructor call!
        """
        self._level = min(MAX_LEVEL, max(MIN_LEVEL, level))
        self._log = logging.getLogger(name)
        self._log.setLevel(self.level)

        # remove handlers (sometimes there are mutliple by default)
        for handler in self._log.handlers:
            self._log.removeHandler(handler)

        # create new handler and formatter
        formatter = ShortLevelFormatter(fmt=fmt, datefmt=datefmt)
        formatter.add_level(notice, 'note')
        stdout_handler = logging.StreamHandler(formatter)
        stdout_handler.setLevel(self.level)
        self._log.addHandler(stdout_handler)

    def debug(self, message, *args, **kwargs):
        self.log(DEBUG, message, *args, **kwargs)

    def info(self, message, *args, **kwargs):
        self.log(INFO, message, *args, **kwargs)

    def notice(self, message, *args, **kwargs):
        self.log(NOTICE, message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        self.log(WARNING, message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        self.log(ERROR, message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        self.log(CRITICAL, message, *args, **kwargs)

    def log(self, level, message, *args, **kwargs):
        if level >= self._level:
            self._log.log(level, message.format(*args), **kwargs)

    def get_level(self):
        """ return int level of this logger """
        return self._level

    def get_level_str(self):
        """ returns :py:func:`logging.getLevelName` on :py:meth:`get_level` """
        return logging.getLevelName(self._level)

    def set_level(self, new_level):
        """ set level given an int or a str

        :arg new_level: int or str (str is converted to lower case)
        :raises: KeyError if new_level is a string that is not in
                 :py:data:`LEVEL_DICT`
        """
        if isstr(new_level):
            self._level = LEVEL_DICT[new_level.lower()]
        else:
            self._level = min(MAX_LEVEL, max(MIN_LEVEL, level))
        self._log.setLevel(self._level)
        for handler in self._log.handlers:
            handler.setLevel(self._level)


def test_short_level_format():
    """ quick test of :py:class:`ShortLevelFormatter` """

    logger = logging.getLogger('logtest')
    logger.setLevel(DEBUG)
    handler = logging.StreamHandler()
    handler.setLevel(DEBUG)
    formatter = ShortLevelFormatter(
        '%(asctime)s:%(msecs)03d %(shortlevel)s| %(msg)s'
        '   [regular levelname=%(levelname)s]',
        datefmt='%H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # 'application' code
    logger.debug('debug message')
    logger.info('info message')
    logger.warn('warn message')
    logger.error('error message')
    logger.critical('critical message')
    logger.log(15, 'unknown level')
    logger.log(NOTSET, 'level not set')

    # add notice level
    notice = (logging.INFO + logging.WARNING)/2
    formatter.add_level(notice, 'note')
    logger.log(notice, 'more important than info but no warning nor error')

    # try if exception formatting still works:
    try:
        logger.info('this is what an exception looks like:')
        impossible_result = 1/0
        logger.critical('just divided 1/0! The result is {0}'
                        .format(impossible_result))
    except ZeroDivisionError:
        logger.exception('1/0 still does not work!', exc_info=True)

    # done
    logger.info('done testing')


if __name__ == '__main__':
    test_short_level_format()
